//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Clans.java                                                       *
//* CLASS    Groups.java. (Groups of agents spatially colocated)              *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     06/06/2015                                                       *
//* REVISION X.0.1.2                                                          *
//* BYTES    00,000                                                           *
//* TIME     15:55                                                            *
//* PURPOSE  This object is the binder that holds the grid cells together.    *
//* METHOD   Each instance of a class is a stand-alone code object.           *
//* DETAILS  A Java document exists that outlines class operation.  Other     *
//*          documentation in the form of programmer notes apply.             *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root.ecology.groups;                          // sub-directory storage.

import java.util.ArrayList;                                   // Java language.
import java.util.Random;                                      //         ditto.

import root.ModelConstraints;                              // program routines.
import root.ecology.Cell;                                  //            ditto.
import root.hominini.Hominid;                              //            ditto.

//
//*****************************************************************************
// Groups

public class Groups implements ModelConstraints
{

  //
  //***************************************************************************
  // Class constants, variables, and objects.

  // Constants.

  // Variables.
  private long rng_Seed = 0L;                    // local copy of the RNG seed.
  private int horizontalLimit = 0;                  // horizontal grid surface.
  private int verticalLimit = 0;                      // vertical grid surface.

  // Objects.
  private Random RNG = null;                 // pointer to standard RNG object.

  public Cell grid [][] = null;   // a copy of the pointer to the terrain grid.

  //***************************************************************************
  /** This is the Terrain (grid) class constructor.*/

  public Groups (long newSeed, Cell theGrid [][], int hLimit, int vLimit)
  {

    this.rng_Seed = newSeed;                       // copy the GLOBAL RNG seed.
    RNG = new Random(this.rng_Seed);    // instantiate an RNG object with seed.

    this.grid = theGrid;                   // local pointer to the GLOBAL grid.

    this.horizontalLimit = hLimit;                        // take a local copy.
    this.verticalLimit = vLimit;                          // take a local copy.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and adds the
      individual hominid (also named by parameter) to the cell's cohort.*/

  public void addCohortIndividual (int h, int v, Hominid thisHominid)
  {

    this.grid[h][v].cohort.add (thisHominid);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method calculates and returns the average cell cohort occupancy.
      If the count of potentially occupied cells is zero, it returns -1.*/

  public int averageCohortSize ()
  {

    ArrayList <Cell> occupiedCells = getOccupiedCells ();
    int countOfOccupiedCells = occupiedCells.size ();

    int countOfOccupiers = 0;
    for (Cell c : occupiedCells)
    {
      countOfOccupiers += c.cohort.size();
    }

    if (countOfOccupiedCells == 0)
      return -1;         // if everyone on the terrain is dead, 0 is the count.
    else
      return countOfOccupiers / countOfOccupiedCells;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns an ArrayList of the Hominids currently occupying
      the grid.*/

  public ArrayList <Hominid> getCellCohorts ()
  {

    ArrayList <Hominid> list = null;

    for (int h = 0; h < this.horizontalLimit; h++)
    {
      for (int v = 0; v < this.verticalLimit; v++)
      {
        if (getCohortCount (h, v) > 0)
          list.addAll (getCohort (h, v));
      }
    }

    return list;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and returns the
      list containing the cell's cohort.*/

  public ArrayList <Hominid> getCohort (int h, int v)
  {

    return this.grid[h][v].cohort;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and returns the
      number of hominids in the cell's cohort arraylist.*/

  public int getCohortCount (int h, int v)
  {

    return this.grid[h][v].cohort.size();

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the count of currently occupied cells in the grid.*/

  public int getNrOfOccupiedCells ()
  {

    int sum = 0;
    for (int h = 0; h < this.horizontalLimit; h++)
    {
      for (int v = 0; v < this.verticalLimit; v++)
      {
        if (getCohortCount (h, v) > 0)
          sum += 1;
      }
    }
    return sum;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns an ArrayList of the cells in the grid that
      are currently occupied.*/

  public ArrayList <Cell> getOccupiedCells ()
  {

    ArrayList <Cell> list = new ArrayList <Cell> ();

    for (int h = 0; h < this.horizontalLimit; h++)
    {
      for (int v = 0; v < this.verticalLimit; v++)
      {
        if (getCohortCount (h, v) > 0)
          list.add ( this.grid [h] [v] );
      }
    }

    return list;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method initializes the terrain subsystem.*/

  public Cell [][]  initializeGroups ()
  {

    for (int h = 0; h < this.horizontalLimit; h++)
    {
      for (int v = 0; v < this.verticalLimit; v++)
      {
        setCohort (h, v, new ArrayList <Hominid> ());   // start an empty list.
      }
    }

    return this.grid;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and removes the
      individual hominid (also named by parameter) from the cell's cohort.*/

  public void removeCohortIndividual (int h, int v, Object thisHominid)
  {

    this.grid[h][v].cohort.remove (thisHominid);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method addresses the cell named by the parameters and sets the
      list containing the cell's cohort.*/

  public void setCohort (int h, int v, ArrayList <Hominid> newCohort)
  {

    this.grid[h][v].cohort = newCohort;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is available to update terrain-related functions.*/

  public synchronized void run ()
  {

    // add code here.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does nothing.*/

  public boolean template ()
  {

    return false;

  } // End method.
  //***************************************************************************

}
