//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Clans.java                                                       *
//* CLASS    Biology.java. (Biology)                                          *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     06/06/2015                                                       *
//* REVISION X.0.1.2                                                          *
//* BYTES    00,000                                                           *
//* TIME     15:55                                                            *
//* PURPOSE  To emulate biological properties and mechanisms.                 *
//* METHOD   By means of software simulation, model proto-humans.             *
//* DETAILS  Created in Java to model proto-humans.                           *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//* COMPILE  javac -g -verbose -deprecation Model.java                        *
//* RUN AS   java -client -Xms1500m -Xmx1500m -Xss1m GenderDominance -cp      *
//*                                                                           *
//*****************************************************************************
//

package root.hominini;                                       // home directory.

import java.util.Random;                                      // Java language.

import root.ModelConstraints;                          // program dependencies.

//
//*****************************************************************************
// Biology

public class Biology implements ModelConstraints
{

  //
  //***************************************************************************
  // Class constants, variables, and objects.

  // Constants.
  // Water constraints, in calories (Citation needed)
  private final float M_HIGH_CAL_NEED = 2450.0F;   // oldest male calorie need.
  private final float M_LOW_CAL_NEED = 950.0F;   // youngest male calorie need.
  private final float F_HIGH_CAL_NEED = 1350.0F; // oldest female calorie need.
  private final float F_LOW_CAL_NEED = 850.0F; // youngest female calorie need.

  // Water constraints, in gallons (Citation needed)
  private final float M_HIGH_WAT_NEED = 0.750F;       // oldest male gal. need.
  private final float M_LOW_WAT_NEED = 0.425F;      // youngest male gal. need.
  private final float F_HIGH_WAT_NEED = 0.700F;     // oldest female gal. need.
  private final float F_LOW_WAT_NEED = 0.400F;    // youngest female gal. need.

  // Weight constraints, in pounds for adults and babies (Citation needed)
  private final float HEALTH_6YR_M_LBS = 15.0F;     // healthy 6yr male weight.
  private final float HEALTH_40YR_M_LBS = 155.0F;  // healthy 40yr male weight.
  private final float HEALTH_6YR_F_LBS = 12.0F;   // healthy 6yr female weight.
  private final float HEALTH_40YR_F_LBS = 120.0F; // health 40yr female weight.
  public  final float BABY_WEIGHT = 6.0F;               // healthy baby weight.

  // Infant loads on mother in calories and gallons (Citation needed)
  private final float BABY_CALORIE_LOAD = 275.0F;// F-only, gest&nursng cal ld.
  private final float BABY_WATER_LOAD = 0.275F;// F-only, gest&nursng water ld.

  // Reproductive constraints (Citation needed)
  private final int   MALE_FERT_AGE = 13; //M-only, minimum age of M fertility.
  private final int   FEM_FERT_AGE = 15; // F-only, minimum age of F fertility.
  public  final int   FEM_NF_CYC_DAYS = 23;  // F-only, nr of days non-fertile.
  public  final int   FEM_F_CYC_DAYS = 5;    // F-only, number of days fertile.
  private final int   GESTATION_END = 252;     // F-only, day # at end of preg.
  private final int   NURSING_END = 1092; //F-only, day # nursing ends (>3yrs).

  // End-of-Life constraints (Citation needed)
  private final float DEATH_LBS_PERCENT = 0.650F;  // death trigger percentage.

  // Variables.
  private int     age = 0;  // in years, delta between birthDate and deathDate.
  private String  sex = "N";      // N=needs reassignment, M=male, or F=female.
  private String  lifeStatus = "N";// N=needs reassignment, D=dead, or A=adult,
                                                       // J=juvenile or B=baby.
  public  String  chromosome = "";    // chromosome that belongs to this agent.
  private int     weaningAge = 0;  // randomly set value in range (see setter).
  private float   waterRequirement = 0.0F;     // a hominid has no requirement.
  private float   caloricRequirement = 0.0F;     // hominid has no requirement.
  private float   weight = 0.0F;          // initially a hominid has no weight.
  private float   fatalWeight = 0.0F;     // initially a hominid has no weight.
  private float   weightPreference = 0.0F;     // a hominid has no preferences.
  private boolean isPregnant = false;                  // F-only, non-pregnant.
  private boolean isNursing = false;                    // F-only, not nursing.
  public  boolean newBirth = false; // initially there are no births to report.
  private int     hominidMaxBirthLimit = 0;    // F-only, limits set by probab.
  private int     lifetimeBirths = 0; // F-only, count of successful birthings.
  private int     dateInFCycle = 0;         // F-only, date in fertility cycle.
  private int     gestationDay = 0;         // F-only, date in gestation cycle.
  private int     nursingDay = 0;             // F-only, date in nursing cycle.
  private boolean matingNotice = false; // F-only, initially no intent to mate.
  private boolean fertile = false;                // F-only, intitially not so.
  private Hominid mObject = null;     // F-only, mated object, father of child.

  private long rng_Seed;                        // this is the GLOBAL RNG seed.

  // Objects.
  public static Hominid self = null;  // pointer to the self available to self.
  private ArtificialGenetics aGenetics = null;    // pointer to the GA methods.

  private static Random RNG = null;          // pointer to standard RNG object.

  //***************************************************************************
  /** This is a class constructor.  It is a default (dummy) structure.*/

  public Biology ()
  {

    // this is intentionally blank.  It is used by the agents started as seed.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** WARNING: this is long method.*/

  /** This is a class constructor.  It makes adults between 15-35 years old.*/

  public Biology (long newSeed, Hominid thisSelf)
  {

    this.rng_Seed = newSeed;                       // copy the GLOBAL RNG seed.
    RNG = new Random(this.rng_Seed);                // an RNG object with seed.

    self = thisSelf;             // complete the self-reference pointer system.

    this.aGenetics = new ArtificialGenetics(RNG.nextInt()); // GA methods link.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the agent's age.*/

  public int getAge()
  {

    return this.age;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the hominid's caloric requirement.*/

  public float getCaloricRequirement ()
  {

    return this.caloricRequirement;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets this hominid's chromosome.*/

  public String getChromosome ()
  {

    return this.chromosome;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the hominid's momentary and current weight.*/

  public float getCurrentWeight ()
  {

    return this.weight;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the hominid's death weight.*/

  public float getFatalWeight ()
  {

    return this.fatalWeight;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets the date in the FCycle (F-only).*/

  public int getFCycleDate ()
  {

    return this.dateInFCycle;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets the F-agents state of fertility.*/

  public boolean getFertile ()
  {

    return this.fertile;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the female's day of gestation.*/

  public int getGestationDay()
  {

    return this.gestationDay;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This calculates and returns the agent's (ideal) preferred weight by
      age, sex, and (for females) state of pregnacy/nursing.  Based on a
      linear relation of age/caloric need from the youngest to the oldest,
      its form is Y = mX + b; where b is calculated and provided for each
      sex.  This is done at agent instantiation, agent birthdays, start of
      gestation (females) and end of nursing (females).*/

  public float getIdealWeight (int age, String sex, int minAge, int maxAge)
  {

    final float CALIB_YEAR = 6.0F;   // used to determine b in y = mx + b form.
    float slope = 0.0F;                    // delta rise over delta run method.
    float b = 0.0F;                    // y-axis offset in year of calibration.

    if (sex.equals("F"))
    { // calculate Y = mX + b.
      slope = (HEALTH_40YR_F_LBS - HEALTH_6YR_F_LBS) / (maxAge - minAge);
      b = HEALTH_6YR_F_LBS - (slope * CALIB_YEAR);
    }

    if (sex.equals("M"))
    { // calculate Y = mX + b.
      slope = (HEALTH_40YR_M_LBS - HEALTH_6YR_M_LBS) / (maxAge - minAge);
      b = HEALTH_6YR_M_LBS - (slope * CALIB_YEAR);
    }

    return (slope * (float) age) + b;                            // y = mx + b.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method evaluates the hominid parameter and according to age and
      sex returns one of three legal life status values: B, J, A.  This
      information may be useful at any time but it is particularly useful
      at the time of death for statistics purposes and during seed-group
      formation when all that is known is age, sex, and weaning age.*/

  public String getLifeGroup ()
  {

    // legal values for Hominid Life Status are:
    // N=needs reassignment, D=dead, A=adult, J=juvenile or B=baby.

    if (this.age < this.weaningAge)
    {
     return "B"; // if hominid is older than weaning age they are at least (J).
    }
    else if ( (this.sex.equals("F") && (this.age < FEM_FERT_AGE) ) ||
              (this.sex.equals("M") && (this.age < MALE_FERT_AGE)) )
      return "J";  // assign as juvenile if weaned but not yet sexually mature.
    else
      return "A";                                // all others assign as adult.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets the hominids count of births (F-only) over a lifetime.*/

  public int getLifetimeBirths ()
  {

    return this.lifetimeBirths;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets the mating notice flag.*/

  public boolean getMatingNotice ()
  {

    return this.matingNotice;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the CONSTANT value for minimum sexual ferility.*/

  public int getFemFertAge ()
  {

    return FEM_FERT_AGE;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the CONSTANT value for minimum sexual ferility.*/

  public int getMaleFertAge ()
  {

    return MALE_FERT_AGE;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets the mObject (mated object) held in storage (F-only).*/

  public Hominid getMObject ()
  {

    return this.mObject; // F-only, mated object, father of child.  Else, null.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns a female's day of nursing.*/

  public int getNursingDay()
  {

    return this.nursingDay;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the status of a female's nursing.*/

  public boolean getNursingStatus()
  {

    return this.isNursing;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets the Hominid Max Birth Limit.*/

  public int getHominidMaxBirthLimit ()
  {

    return this.hominidMaxBirthLimit;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the hominid's life status (B, J, A, D, N, etc.).*/

  public String getHominidLifeStatus ()
  {

    return this.lifeStatus;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the hominid's pregnancy truth state.*/

  public boolean getPregnancyStatus ()
  {

    return this.isPregnant;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the agent's sex.*/

  public String getSex ()
  {

    return this.sex;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the hominid's water requirement.*/

  public float getWaterRequirement ()
  {

    return this.waterRequirement;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This sets the hominid's water requirements.*/

  public void setWaterRequirement (float newWaterRequirement)
  {

    this.waterRequirement = newWaterRequirement;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the age when weaned.*/

  public int getWeaningAge ()
  {

    return this.weaningAge;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns the hominid's preferred (or gender & age based) weight.*/

  public float getWeightPreference ()
  {

    return this.weightPreference;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This increments the gestation term by one day.*/

  public void incrementGestationDay ()
  {

    int tempCount = getGestationDay();
    setGestationDay (++tempCount);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This increments a nursing term by one day.*/

  public void incrementNursingDay ()
  {

    int tempCount = this.getNursingDay();
    this.setNursingDay (++tempCount);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This calculates and returns a momentary caloric requirement for the
      agent by age, gender, and (for females) state of pregnacy/nursing.
      Based on a linear relation of age/caloric need from the youngest to the
      oldest, its form is Y = mX + b; where b is the minimum required by the
      agent's gender.*/

  public float momentaryCaloricReq (int age, String sex)
  {

    return momentaryCaloricReq (age, sex, self.MINIMUM_AGE, self.MAXIMUM_AGE);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This calculates and returns a momentary caloric requirement for the
      agent by age, gender, and (for females) state of pregnacy/nursing.
      Based on a linear relation of age/caloric need from the youngest to the
      oldest, its form is Y = mX + b; where b is the minimum required by the
      agent's gender.*/

  public float momentaryCaloricReq (int age, String sex,
                                    int minAge, int maxAge)
  {

    final float CALIB_YEAR = 6.0F;   // used to determine b in y = mx + b form.
    float slope = 0.0F;                    // delta rise over delta run method.
    float b = 0.0F;                    // y-axis offset in year of calibration.

    if (sex.equals("F"))
    { // calculate Y = mX + b.
      slope = (F_HIGH_CAL_NEED - F_LOW_CAL_NEED) / (maxAge - minAge);
      b = F_LOW_CAL_NEED - (slope * CALIB_YEAR);
    }

    if (sex.equals("M"))
    { // calculate Y = mX + b.
      slope = (M_HIGH_CAL_NEED - M_LOW_CAL_NEED) / (maxAge - minAge);
      b = M_LOW_CAL_NEED - (slope * CALIB_YEAR);
    }

    // for females this is where we adjust caloric load for pregnancy/nursing.
    if (sex.equals("F") && (getPregnancyStatus() || getNursingStatus()))
      return (slope * ((float) age)) + b + BABY_CALORIE_LOAD;   // increase up.
    else
      return (slope * (float) age) + b;                          // y = mx + b.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This calculates and returns a momentary water requirement for the
      agent by age, gender, and (for females) state of pregnacy/nursing.
      Based on a linear relation of age/caloric need from the youngest to the
      oldest, its form is Y = mX + b; where b is the minimum required by the
      agent's gender.*/

  public float momentaryWaterReq (int age, String sex)
  {

    return momentaryWaterReq (age, sex, self.MINIMUM_AGE, self.MAXIMUM_AGE);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This calculates and returns a momentary water requirement for the
      agent by age, gender, and (for females) state of pregnacy/nursing.
      Based on a linear relation of age/caloric need from the youngest to the
      oldest, its form is Y = mX + b; where b is the minimum required by the
      agent's gender.*/

  public float momentaryWaterReq (int age, String sex,
                                  int minAge, int maxAge)
  {

    final float CALIB_YEAR = 6.0F;   // used to determine b in y = mx + b form.
    float slope = 0.0F;                    // delta rise over delta run method.
    float b = 0.0F;                    // y-axis offset in year of calibration.

    if (sex.equals("F"))
    { // calculate Y = mX + b.
      slope = (F_HIGH_WAT_NEED - F_LOW_WAT_NEED) / (maxAge - minAge);
      b = F_LOW_WAT_NEED - (slope * CALIB_YEAR);
    }

    if (sex.equals("M"))
    { // calculate Y = mX + b.
      slope = (M_HIGH_WAT_NEED - M_LOW_WAT_NEED) / (maxAge - minAge);
      b = M_LOW_WAT_NEED - (slope * CALIB_YEAR);
    }

    // for females this is where we adjust water loading for pregnancy/nursing.
    if (sex.equals("F") &&
       (getPregnancyStatus() || getNursingStatus()))
      return (slope * (float) age) + b + BABY_WATER_LOAD;       // increase up.
    else
      return (slope * (float) age) + b;                          // y = mx + b.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This sets the agent's age.*/

  public void setAge (int newAge)
  {

    this.age = newAge;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This sets the hominid's caloric requirements.*/

  public void setCaloricRequirement (float newCaloricRequirement)
  {

    this.caloricRequirement = newCaloricRequirement;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this HGs chromosome.*/

  public void setChromosome (String newValue)
  {

    this.chromosome = newValue;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This sets the hominid's momentary and current weight.*/

  public void setCurrentWeight (float newWeight)
  {

    this.weight = newWeight;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This gene influences the capacity of the bearer to survive extreme
      weight loss, i.e., to survive otherwise withering weight loss due to
      disease and or malnutrition.  ROBUST_METABOLISM expresses for both
      sexes (genders).  At birth, the value of this gene is set by random
      Gaussian distribution over the integer interval 0 to 255, inclusive.
      The bearers ultimate mortality weight is based in part on the base
      value of the gene and some calculation.  The genetic advantage ranges
      from 0.0 to about 5.1 pounds of additional, survivable weight loss
      taken arbitrarily from within a range of 35-50%of survivable total
      body mass loss projected by forensic practitioner.  A system standard
      mortality weight of 1.00 � 0.35 or 0.65 (65%) of standard bearer body
      mass set by sex and age is always available to the bearer.  The bearer
      of this gene may enjoy a slight advantage.  This genetic trait is
      phenotypic.*/

  public void setFatalWeight ()
  {

    setFatalWeight (getChromosome(), self.ROBUST_METABOLISM_GENE);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This gene influences the capacity of the bearer to survive extreme
      weight loss, i.e., to survive otherwise withering weight loss due to
      disease and or malnutrition.  ROBUST_METABOLISM expresses for both
      sexes (genders).  At birth, the value of this gene is set by random
      Gaussian distribution over the integer interval 0 to 255, inclusive.
      In operation however, this gene operates on geneValues 128-255 only.
      Values 0-127, adhere to the standard mortality ration (65%) described
      below. For juveniles and adults there is genetic advantage.  For adults
      this can range from 0.0 to about 5.1 pounds of additional, survivable
      weight loss.  This is derived survivable total body mass projections
      cited by forensic practitioners.  Juveniles experience a more
      constrained advantage having a maximum "arbitrary" superiority of 10%
      across the entire 128-355 range.  The system standard survivable
      mortality weight is 0.65 (65%) of bearer preferred body weight.  The
      bearer of this gene may enjoy a slight adaptive advantage over others.
      This genetic trait is phenotypic.*/

  public void setFatalWeight (String chromosome, int gene)
  {

    final float ARBITRARY_ADVANTAGE = 127.0F;   // arbitrary advantage of gene.
    float mortalityWeight = getWeightPreference() * DEATH_LBS_PERCENT;

    int geneValue = 0;                // ranges from 0 - 255 in the 8-bit gene.
    geneValue = this.aGenetics.getGeneValue(chromosome, gene);

    // outcomes in range ~0.0-5.1 pounds advantage (i.e., small) are possible.
    if (geneValue == 0) geneValue = 1;            // safety for zero division.

    float geneticAdvantage = (float) geneValue / ARBITRARY_ADVANTAGE;
    if (geneticAdvantage > 1.0F)
    {

      // limit "J" juveniles to 10% advantage but allow "A" adults full advan.
      if (getLifeGroup().equals("J"))
        this.fatalWeight = mortalityWeight - (mortalityWeight * 0.10F); // 10%.
      else
        this.fatalWeight = mortalityWeight - geneticAdvantage;   // full advan.

    }
    else
      this.fatalWeight = mortalityWeight;          // all others yet undefined.

    // SPECIAL CASE:  Babies' fatal weight always rests with mother.
    if (getHominidLifeStatus().equals("B")) this.fatalWeight = BABY_WEIGHT;

    // Death as a result of starvation. Madea, Burkhard. Forensic pathology
    // reviews. Humana Press, 2005. 3-23.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the date in the FCycle (F-only).*/

  public void setFCycleDate (int newFCycleDate)
  {

    this.dateInFCycle = newFCycleDate;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the F-agents state of fertility.*/

  public void setFertile (boolean newFertileState)
  {

    this.fertile = newFertileState;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This sets the female's day of gestation.*/

  public void setGestationDay (int newDay)
  {

    this.gestationDay = newDay;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the hominids count of births (F-only) over a lifetime.*/

  public void setLifetimeBirths (int newCount)
  {

    this.lifetimeBirths = newCount;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the mating notice flag.*/

  public void setMatingNotice (boolean newMatingNotice)
  {

    this.matingNotice = newMatingNotice;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the mObject (mated object) held in storage (F-only).*/

  public void setMObject (Hominid newMObject)
  {

    this.mObject = newMObject; // F-only, mated object, father of child.
                               // Else, null.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This returns a notice of a new birth for external pick-up.*/

  public boolean getNewBirthNotice()
  {

    return this.newBirth;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This sets a notice of a new birth for external pick-up.*/

  public void setNewBirthNotice (boolean recordBirth)
  {

    this.newBirth = recordBirth;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This sets a female's day of nursing.*/

  public void setNursingDay (int newDay)
  {

    this.nursingDay = newDay;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This sets the status of a female's nursing.*/

  public void setNursingStatus (boolean newStatus)
  {

    this.isNursing = newStatus;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets the Hominid Max Birth Limit.*/

  public void setHominidMaxBirthLimit (int newLimit)
  {

    this.hominidMaxBirthLimit = newLimit;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This sets the hominid's life status: D=Dead, A=Adult, B=Baby.*/

  public void setHominidLifeStatus (String newStatus)
  {

    this.lifeStatus = newStatus;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This sets the hominid's pregnancy truth state.*/

  public void setPregnancyStatus (boolean thisPregnancy)
  {

    this.isPregnant = thisPregnancy;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This sets the agent's sex.*/

  public void setSex (String newValue)
  {

    this.sex = newValue;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets this individuals weaning age in the range of 4 - 6
      years old, inclusive.*/

  public void setWeaningAge ()
  {

    this.weaningAge = MIN_WEANING_AGE + RNG.nextInt (RANGE_OF_EARLY_WEANING);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This set the hominid's preferred (or gender & age based) weight.*/

  public void setWeightPreference (float newWeightPreference)
  {

    this.weightPreference = newWeightPreference;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /**REPRODUCTIVE CONTROLS:  checks reproduction & fertility related controls.
     Factors under consideration include:

     ESTRUS:  update or establish female fertility cycle.
     NURSING: check if this hominid is nursing.
     PREGNANCY: check if this hominid is pregnant.*/

  public void updateReproductiveStatus (int age, Socioecology kin,
                                        int minAge, int maxAge,
                                        String chromosome, int gene)
  {

    // ------------------------------------------------------------------------
    // ESTRUS:  update or establish female fertility cycle.

    if (age >= FEM_FERT_AGE)
    {
      int dateInCycle = getFCycleDate();      // day in F cycle 0-27 inclusive.

      if ( dateInCycle < ( FEM_NF_CYC_DAYS + FEM_F_CYC_DAYS - 1 )) // tests 27.
        setFCycleDate( ++dateInCycle);                 // increment date clock.
      else
        setFCycleDate(0);                                  // reset date clock.

      if ((getFCycleDate() % FEM_NF_CYC_DAYS) == 0)
        setFertile(true);                             // entering fertile time.

      if (getFCycleDate() == 0)
        setFertile(false);                             // leaving fertile time.

      // ----------------------------------------------------------------------
      // NURSING: check if this hominid is nursing.

      if (getNursingStatus())
      {
        setFCycleDate(0);                        // inaccurate prolactin delay.
        incrementNursingDay();                           // increment the term.

        int c = kin.childrenCount () - 1;    // only the last child is in play.
        // check if has no child, nursing past nursing end, or child is dead.
        if (!kin.hasChildren() || (getNursingDay() >= NURSING_END)||
          (kin.getChild(c).bio.getHominidLifeStatus().equals("D")))
        {
          setNursingDay(0);                                        // set term.
          setNursingStatus(false);              // set the hominid status flag.
          // recalculate the metabolic rate needed for pregnancy to succeed.
          setWeightPreference(getIdealWeight (getAge(), getSex(),
                              minAge, maxAge));
          setCaloricRequirement (momentaryCaloricReq(getAge(), getSex(),
                                 minAge, maxAge));
          setWaterRequirement (momentaryWaterReq(getAge(), getSex(),
                               minAge, maxAge));
          // survivable minimum living weight - uses gene 2.
          setFatalWeight(chromosome, gene);
        }
      }

      // ----------------------------------------------------------------------
      // PREGNANCY: check if this hominid is pregnant.

      if (getPregnancyStatus())           // check if this hominid is pregnant.
      {
        setFCycleDate(0);                        // inaccurate prolactin delay.
        incrementGestationDay();                         // increment the term.

        if (getGestationDay() > GESTATION_END)         // check if end of term.
        { // this was a birth.
          setNewBirthNotice(true);                   // record this as a birth.
          setPregnancyStatus(false);            // set the hominid status flag.
          setNursingStatus(true);               // set the hominid status flag.
          setGestationDay(0);                                      // set term.
          setNursingDay(0);                                        // set term.
        }
      }

      // ----------------------------------------------------------------------
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does nothing.*/

  public boolean template ()
  {

    return false;

  } // End method.
  //***************************************************************************

}