//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Clans.java                                                       *
//* CLASS    HAL (Heuristic, Algorithmic, Learning)                           *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     06/06/2015                                                       *
//* REVISION X.0.1.2                                                          *
//* BYTES    00,000                                                           *
//* TIME     15:55                                                            *
//* PURPOSE  This class provides several primitive, prioritized, drive        *
//*          reduction schemes.  First, it assumes the hominid can perceive   *
//*          sensations that have been previously paired (a priori) with a    *
//*          drive reducing response.  This is similar to a trigger stimulus  *
//*          setting off a fixed-action-pattern.  Second, this facilitates    *
//*          the Hominid's action taking (responses) within the constraints   *
//*          of a weighted prioritization scheme.  The goal is the            *
//*          emulation of a set of complex, overt, drive reduction            *
//*          behaviors.                                                       *
//* METHOD   Each instance of a class is a stand-alone code object.           *
//* DETAILS  A Java document exists that outlines class operation.  Other     *
//*          documentation in the form of programmer notes apply.             *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root.hominini;                                // sub-directory storage.

import java.awt.Point;                                        // Java language.
import java.util.ArrayList;                                   //         ditto.
import java.util.Random;                                      //         ditto.

import root.hominini.Hominid;                              // program routines.
import root.ecology.Ecology;                               //            ditto.
import root.ModelConstraints;                              //            ditto.

//
//*****************************************************************************
// HAL

public class HAL implements ModelConstraints
{

  //***************************************************************************
  // Class variables and objects.

  // Constants.

  // Variables.
  private static long rng_Seed = 0L;            // this is the GLOBAL RNG seed.

  public static int reqX = 0;   // the movement transfer request register in X.
  public static int reqY = 0;   // the movement transfer request register in Y.
  private ArrayList <Point> bresenhams;  // a straight line between two points.

  private static int hLimit = 0;        // the user's map horizontal dimension.
  private static int vLimit = 0;          // the user's map vertical dimension.

  private static int drivePriorities = 0;    // count of subEpoch event drives.

  // Object.
  private static Random RNG = new Random();           // a standard RNG object.
  private static Random giftingRNG = null;   // pointer to Alliance RNG - ONLY.

  //***************************************************************************
  /** This is the class constructor.*/

  public HAL (long seed, int hStartLoc, int vStartLoc,
                         int horizontalLimit, int verticalLimit)
  {

    this.rng_Seed = seed;                          // copy the GLOBAL RNG seed.
    RNG = new Random(seed);             // instantiate an RNG object with seed.

    giftingRNG = new Random(seed);          // an RNG object with a fixed seed.

    this.reqX = hStartLoc;                                   // start register.
    this.reqY = vStartLoc;                                   // start register.

    // the constraints of the user's map.
    hLimit = horizontalLimit;
    vLimit = verticalLimit;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This uses Bresenhams algorithm to find the X and Y coordinates
      associated with a straight line between Points A and B on a flat plane.
      Assume you are at Point A and you want to get to Point B.  The result
      of calling the bresenhamsAlgorithm is an inclusive set of points along
      the line with that Point set put into the bresenhams ArrayList.*/

  private void bresenhamsAlgorithm ( Point A, Point B )
  {

    // http://en.wikipedia.org/wiki/Bresenham's_line_algorithm
    this.bresenhams = new ArrayList < Point > (); // a container for this line.

    int xStart = (int) A.getX ();
    int yStart = (int) A.getY ();
    int xEnd = (int) B.getX ();
    int yEnd = (int) B.getY ();
    int tempVarX = 0;
    int tempVarY = 0;
    int error1 = 0;
    int error2 = 0;

    int deltaX = Math.abs ( xEnd - xStart );
    int deltaY = Math.abs ( yEnd - yStart );
    Point p = null;                   // this is the next point as it is found.

    if ( xStart < xEnd )
      tempVarX = 1;
    else
      tempVarX = -1;

    if ( yStart < yEnd )
      tempVarY = 1;
    else
      tempVarY = -1;

    error1 = deltaX - deltaY;
    do
    {
      p = new Point (xStart, yStart);
      this.bresenhams.add ( p );

      if ( ( xStart == xEnd ) && ( yStart == yEnd ) )
      {
        p = null;                 // return the memory to the operating system.
        return;
      }

      error2 = 2 * error1;
      if ( error2 > - deltaY )
      {
        error1 = error1 - deltaY;
        xStart = xStart + tempVarX;
      }

      if ( ( xStart == xEnd ) && ( yStart == yEnd ) )
      {
        p = new Point (xStart, yStart);
        this.bresenhams.add ( p );
        p = null;                 // return the memory to the operating system.
        return;
      }

      if ( error2 < deltaX )
      {
        error1 = error1 + deltaX;
        yStart = yStart + tempVarY;
      }
    }
    while ( true );

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns a filtered String representation of binary switches.
      The switches are the Boolean values of affect triggers encoded in sMask
      during Hominid reduceDrives execution.  Thesebits were determined to be
      necessary but require prioritization before they can be activated. The
      scheme here that filters whether a trigger is to be prioritized
      involves reading the sMask from left to right (0 to n) and keeping
      only the topPriority bits and discarding (zeroing) the rest.  A result
      is a string keeping the topPrioritys and zeroing the remainder.  The
      position of each bit is important so whatever placement was given will
      be returned.*/

  public String drivePrioritization (String sMask, int topPriorities)
  {

    drivePriorities = 0;         // reset the counter of subEpoch event drives.
    if (topPriorities <= 0)
    {
      System.out.println ("ERROR mode: HAL.drivePrioritization");
      System.exit(0);
    }

    String result = new String ("");
    int maskLength = sMask.length();

    int priorityCounter = 0;
    int pointer = 0;
    do
    {
      if (sMask.substring(pointer, (pointer + 1)).equals("0"))
      {
        result += "0";                                       // Boolean false.
      }
      else // must be a "1"
      {
        result += "1";                                         // Boolean true.
        priorityCounter++;               // tally the count of priority events.
        drivePriorities++;                    // subEpoch event driver counter.
      }
    } while ((++pointer < maskLength) && (priorityCounter < topPriorities));

    for (int i = pointer; i < sMask.length(); i++)
    {
      result += "0";
    }

    return result;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** WARNING: this is long method.

      This method emulates working memory behavioral control activities
      related to perceptual cognitive dissonance.  Here, when a sensory
      apprehension is categorically sufficient but violates existing
      attitudes, a (re)evaluation of the attitude takes place.  Over time,
      it is possible the strength (weight) of the attitude "may" be changed.
      You wouldn't be here if your mental model was sufficient.*/

  public Hominid evalCogDissonance (Hominid self,
                                    Ecology ecoSys, String stimuli)
  {

    int h = self.getCurrentLoc().x;                  // the current X location.
    int v = self.getCurrentLoc().y;                  // the current Y location.

    // this dissonance pertains to the contradictory cognition that occurs
    // when immediate sensory apprehension of a nutritionally sufficient
    // but non-preferential prey food is encountered.
    if ( self.getPreferredPrey().equals( ("none") ) )
      self.setPreferredPrey("fig");
    if ( stimuli.equals( "Prey" ) &&
         !ecoSys.getForageOption(h, v).equals ("none") )
    {
      int timesSatisfied = self.getTimesFoodSatisfied(); // weight of attitude.
      String realPrey = ecoSys.getForageOption(h, v);    // recognize the prey.
      String lastPrey = self.getRememberedPrey();   // the previously had prey.

      // if the prey encountered this time is the same as that encountered
      // last time then, adjust the weighting of the associated attitude.
      if (lastPrey.equals(realPrey))
      {
        timesSatisfied++;           // increase the perception of satisfaction.
        // has this prey satisfied for "long enough" to make it preferential.
        if (timesSatisfied > PREY_ATTITUDE_CHANGE_THRESHOLD)
        {
          // perception has sufficiently overwhelmed attitude.
          self.setPreferredPrey(realPrey);     // change focus of the attitude.
          timesSatisfied = 0;      // reset-restart attitude ratchet mechanism.
          // DIAGNOSTIC.
          //System.out.println("Algorithmics.evalCogDissonance() resolution.");
        }
      }
      else
      { // else decrease the perception of satisfaction.
        timesSatisfied--;           // decrease the perception of satisfaction.
        // in this case,
        // de-value the perception but disallow it from becomin a disgust.
        if (timesSatisfied < 0 )
          timesSatisfied = 0;
      }
      // inform attitude and transfer to LTM the perceived utility stimulus.
      self.setTimesFoodSatisfies(timesSatisfied);    // update attitude weight.
      self.setRememberedPrey(realPrey);                     // move STM to LTM.
    }

    // dissonance pertains to what?
    if (stimuli.equals( "stub" ))
    {
      //add code here as needed.
    }

    return self;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns a point along a Bresenham's line extending between
      the two points meHere and inWater, inclusive.  It is assumed the second
      point, inWater, is in the "water" biome.  The point returned is the
      first point just outside the "water" biome (thus, at the waters edge).
      If there are no non-"water" biome points along the line (i.e., all
      points are in the "water" biome), the point of origin (meHere), a point
      logically also in the "water," will be returned.*/

  public Point findWatersEdge(Ecology ecoSys, Point meHere, Point inWater)
  {

    Point p = null;
    bresenhamsAlgorithm ( meHere, inWater );

    int i = (this.bresenhams.size() - 1); // start with the point in the water.
    do
    {

      // if this point is in the water, move to the next one.
      if (!ecoSys.getBiomeName (this.bresenhams.get(i).x,
                                this.bresenhams.get(i).y).equals ("water"))
      {
        p = this.bresenhams.get (i);       // you should be out of the "water."
      }

    } while ((p == null) && (--i >= 0));

    if (p == null)
      return inWater;
    else
      return p;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns a point along a Bresenham's line extending between
      the two points meHere and inWater, inclusive.  It is assumed the second
      point, in Water, is in the "water" biome.  The point returned is the
      last point in the "water" biome. If there are no non-"water" biome
      points along the line (i.e., all points are in the "water" biome), the
      point of origin (meHere), a point logically also in the "water," will
      be returned.*/

  public Point findShallowWater(Ecology ecoSys, Point meHere, Point inWater)
  {

    Point p = null;
    bresenhamsAlgorithm ( meHere, inWater );

    int i = (this.bresenhams.size() - 1); // start with the point in the water.
    do
    {

      // if this point is in the water, move to the next one.
      if ( !ecoSys.getBiomeName (this.bresenhams.get(i).x,
                                 this.bresenhams.get(i).y).equals ("water") )
      {
        p = this.bresenhams.get (i + 1);  // go back one cell into the "water."
      }

    } while ((p == null) && (--i >= 0));

    if (p == null)
      return inWater;
    else
      return p;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the current count of subEpoch event drives.*/

  public int getDrivePriorities()
  {

    return drivePriorities;              // the count of subEpoch event drives.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets the parameter reqX.*/

  public int getReqX()
  {

    if (this.reqX < 0)
    {
      System.out.println("Error mode: Heuristics.getReqX() - 0");
      System.out.println("            Did you reset/adjust ModelConstraints?");
    }
    if (this.reqX > (hLimit - 1))
    {
      System.out.println("Error mode: Heuristics.getReqX() - max");
      System.out.println("            Did you reset/adjust ModelConstraints?");
    }

    return this.reqX;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets the parameter reqY.*/

  public int getReqY()
  {

    if (this.reqY < 0)
    {
      System.out.println("Error mode: Heuristics.getReqY() - 0");
      System.out.println("            Did you reset/adjust ModelConstraints?");
    }
    if (this.reqY > (vLimit - 1))
    {
      System.out.println("Error mode: Heuristics.getReqY() - max");
      System.out.println("            Did you reset/adjust ModelConstraints?");
    }

    return this.reqY;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This constrains horizontal movement onto a tordial surface. There are
      three possible states: too far left, too far right, or on grid.*/

  public int hBounds (int possibleH)
  {

    if (possibleH < 0)                                         // too far left.
      return hLimit + possibleH;                   // circle in from the right.

    if (possibleH < hLimit)                          // if less than the limit.
      return possibleH;                            // no problem, stay on grid.

    if (possibleH >= hLimit)                               // if too far right.
      return possibleH - hLimit;                    // circle in from the left.

    System.out.println ("ERROR MODE: HAL.hBounds testing " + possibleH);
    return -42;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method assigns the benefactors location in XY to the transfer
      registers.*/

  public void moveForAlliance (Hominid benefactor)
  {

    this.reqX = benefactor.getCurrentLoc().x;    // assign to the transfer reg.
    this.reqY = benefactor.getCurrentLoc().y;    // assign to the transfer reg.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method probabilistically chooses to move the agent from the
      current cell to an adjoining cell (or not).  The resultant XY
      "move to" location is put in the transfer registers.*/

  public void moveForDispersal (Hominid self, Ecology ecoSys)
  {

    int h = self.getCurrentLoc().x;                 // where self is in X-axis.
    int v = self.getCurrentLoc().y;                 // where self is in y-axis.
    setCurrentXY(h, v); // assume stay in the current location, load registers.
    int testH = 0;                       // a value on the X axis to be tested.
    int testV = 0;                       // a value on the Y axis to be tested.

    // choose randomly, generating a step in some direction within limits.
    do
    {
      // choose randomly, propose a step but don't step into the water.
      testH = hBounds((RNG.nextBoolean() ?
                      (RNG.nextBoolean() ? --h : h) :
                      (RNG.nextBoolean() ? h : ++h)));

      testV = vBounds((RNG.nextBoolean() ?
                      (RNG.nextBoolean() ? --v : v) :
                      (RNG.nextBoolean() ? v : ++v)));
    } while (ecoSys.getBiomeName (testH, testV).equals("water"));

    this.reqX = testH;                           // assign to the transfer reg.
    this.reqY = testV;                           // assign to the transfer reg.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method examines the cells adjacent to the current cell.  The
      goal is to identify a cell that satisfies an arbitrary goal.  In
      random order and constrained by the cardinality of the set enumerated
      by count of cells in the Moore neighborhood around this cell (a Moore
      distance set by parameter), the cells are randomly queried.  This
      is is a first fit algorithm.  The first cell that satisfies the query
      will be selected for recommendation.  If no cell is recommended the
      Class request registers (reqX and Y) will not changed.  If there is a
      recommendation, the resultant XY "move to" combination location will
      be put into the transfer registers.*/

  public boolean moveForFear (Hominid self, Ecology ecoSys, int distance)
  {

    int h = self.getCurrentLoc().x;                 // where self is in X-axis.
    int v = self.getCurrentLoc().y;                 // where self is in y-axis.
    setCurrentXY(h, v); // assume stay in the current location, load registers.

    boolean done = false;                    // to begin, the work is not done.
    boolean found = false;              // to begin, the solution is not found.

    int hDelta = 0;      // a distance from the location of self on the X axis.
    int vDelta = 0;      // a distance from the location of self on the Y axis.
    int testH = 0;                       // a value on the X axis to be tested.
    int testV = 0;                       // a value on the Y axis to be tested.

    int side = (2 * distance) + 1; // length of the search box side, an odd nr.
    int cardinality = (side * side) - 1;  // cardinality of adjacent Moore set.
    int counter = 0;       // a counter to compare to the cardinality variable.
    do
    {
      hDelta = ((RNG.nextInt(distance) + 1) * ((RNG.nextBoolean())? +1 : -1));
      vDelta = ((RNG.nextInt(distance) + 1) * ((RNG.nextBoolean())? +1 : -1));
      testH = h + hDelta;
      testV = v + vDelta;
      testH = hBounds ( h + hDelta );     // on a torodial grid so wrap-around.
      testV = vBounds ( v + vDelta );     // on a torodial grid so wrap-around.

      // ------------------------------- QUERY --------------------------------
      found = !ecoSys.getBiomeName(testH, testV).equals("water");
      // ------------------------------- QUERY --------------------------------

      done = ( ( ++counter < cardinality ) ? false : true );
    }
    while ( !done && !found );

    if ( found )
    {
      this.reqX = testH;                         // assign to the transfer reg.
      this.reqY = testV;                         // assign to the transfer reg.
    }

    return found;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method directs the search for a food in cells adjacent to the
      current but at a Moore distance.  First, a search is made with a
      preferred prey in mind.  If the preferred prey food is not available
      then, a final search is made based solely on an expected magnitude
      of potential foraging success.*/

  public void moveForHunger (Hominid self, Ecology ecoSys, int distance)
  {

    boolean foundPrey = false;   // assume the preferred prey is not available.

    // sometimes a seed group will be started on a cell marked by a "none"
    // foraging option.  This is a setup artifact and non-plausible prey
    // preference.  Test for "none" and force such seed group member(s) to
    // search by probability.  Subsequent movement for hunger should thus
    // reinforce an new preference.
    if (!self.getPreferredPrey().equals("none"))
      foundPrey = searchByFoodPreference (self, ecoSys, distance);   // search.
    else
      self.setPreferredPrey("fig");    // seed group member needs a valid prey.

    // if preferred prey was not found find some other based on better chances.
    if (! foundPrey ) searchByFoodProbability (self, ecoSys, distance);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method examines the cells adjacent to the current cell.  The
      goal is to identify a cell that satisfies an arbitrary goal.  In
      random order and constrained by the cardinality of the set enumerated
      by count of cells in the Moore neighborhood around this cell (a Moore
      distance set by parameter), the cells are randomly queried.  This
      is is a first fit algorithm.  The first cell that satisfies the query
      will be selected for recommendation.  If no cell is recommended the
      Class request registers (reqX and Y) will not changed.  If there is a
      recommendation, the resultant XY "move to" combination location will
      be put into the transfer registers.*/

  public void moveForSocialGroup (Hominid self, Ecology ecoSys, int distance)
  {

    int h = self.getCurrentLoc().x;                 // where self is in X-axis.
    int v = self.getCurrentLoc().y;                 // where self is in y-axis.
    setCurrentXY(h, v); // assume stay in the current location, load registers.

    int localGrpSize = ecoSys.getCohortCount(h, v);          // local grp size.
    int remoteGrpSize = 0;                           // nr in the remote group.

    boolean done = false;                    // to begin, the work is not done.
    boolean found = false;              // to begin, the solution is not found.

    int hDelta = 0;      // a distance from the location of self on the X axis.
    int vDelta = 0;      // a distance from the location of self on the Y axis.
    int testH = 0;                       // a value on the X axis to be tested.
    int testV = 0;                       // a value on the Y axis to be tested.

    int side = (2 * distance) + 1; // length of the search box side, an odd nr.
    int cardinality = (side * side) - 1;  // cardinality of adjacent Moore set.
    int counter = 0;       // a counter to compare to the cardinality variable.

    do
    {
      hDelta = ((RNG.nextInt(distance) + 1) * ((RNG.nextBoolean())? +1 : -1));
      vDelta = ((RNG.nextInt(distance) + 1) * ((RNG.nextBoolean())? +1 : -1));
      testH = hBounds ( h + hDelta );     // on a torodial grid so wrap-around.
      testV = vBounds ( v + vDelta );     // on a torodial grid so wrap-around.

      // ------------------------------- QUERY --------------------------------
      remoteGrpSize = ecoSys.getCohortCount(testH, testV);   // poor knowledge.
      found = ( remoteGrpSize > localGrpSize );
      // ------------------------------- QUERY --------------------------------

      done = ( ( ++counter < cardinality ) ? false : true );
    }
    while ( !done && !found );

    if ( found )
    {
     // test if is "water" then constrain movement if it is "into" the water.
      if (ecoSys.getBiomeName (testH, testV).equals("water"))
      {
        Point p = new Point (testH, testV);       // a point next to the water.
        p = findWatersEdge (ecoSys, self.getCurrentLoc(), p);  // water's edge.
        this.reqX = p.x;                         // assign to the transfer reg.
        this.reqY = p.y;                         // assign to the transfer reg.
      }
      else
      { // was tested and found not to be water so just move.
        this.reqX = testH;                       // assign to the transfer reg.
        this.reqY = testV;                       // assign to the transfer reg.
      }
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method probabilistically chooses to move the agent from the
      current cell to an adjoining cell (or not).  Since there is no know way
      to know if the temperatute in an adjacent cell will necessarily be more
      comfortable, the selection process is entirely random but does not
      allow a step into the water. The resultant XY "move to" location is
      put in the transfer registers.*/

  public void moveForTemp (Hominid self, Ecology ecoSys)
  {

    final float HIGH_THRESHOLD = 0.90F;         // establish a LOW probability.

    int h = self.getCurrentLoc().x;                 // where self is in X-axis.
    int v = self.getCurrentLoc().y;                 // where self is in y-axis.
    setCurrentXY(h, v); // assume stay in the current location, load registers.
    int testH = 0;                       // a value on the X axis to be tested.
    int testV = 0;                       // a value on the Y axis to be tested.

    if (RNG.nextFloat() > HIGH_THRESHOLD)  // however according to probability:
    {
      do
      {
        // choose randomly, propose a step but don't step into the water.
        testH = hBounds((RNG.nextBoolean() ?
                        (RNG.nextBoolean() ? --h : h) :
                        (RNG.nextBoolean() ? h : ++h)));

        testV = vBounds((RNG.nextBoolean() ?
                        (RNG.nextBoolean() ? --v : v) :
                        (RNG.nextBoolean() ? v : ++v)));
      } while (ecoSys.getBiomeName (testH, testV).equals("water"));

      this.reqX = testH;                         // assign to the transfer reg.
      this.reqY = testV;                         // assign to the transfer reg.
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method examines the cells adjacent to the current cell.  The
      goal is to identify a cell that satisfies an arbitrary goal.  In
      random order and constrained by the cardinality of the set enumerated
      by count of cells in the Moore neighborhood around this cell (a Moore
      distance set by parameter), the cells are randomly queried.  This
      is is a first fit algorithm.  The first cell that satisfies the query
      will be selected for recommendation.  If no cell is recommended the
      Class request registers (reqX and Y) will not changed.  If there is a
      recommendation, the resultant XY "move to" combination location will
      be put into the transfer registers.*/

  public void moveForThirst (Hominid self, Ecology ecoSys, int distance)
  {

    int h = self.getCurrentLoc().x;                 // where self is in X-axis.
    int v = self.getCurrentLoc().y;                 // where self is in y-axis.
    setCurrentXY(h, v); // assume stay in the current location, load registers.

    boolean done = false;                    // to begin, the work is not done.
    boolean found = false;              // to begin, the solution is not found.

    int hDelta = 0;      // a distance from the location of self on the X axis.
    int vDelta = 0;      // a distance from the location of self on the Y axis.
    int testH = 0;                       // a value on the X axis to be tested.
    int testV = 0;                       // a value on the Y axis to be tested.

    int side = (2 * distance) + 1; // length of the search box side, an odd nr.
    int cardinality = (side * side) - 1;  // cardinality of adjacent Moore set.
    int counter = 0;       // a counter to compare to the cardinality variable.
    do
    {
      hDelta = ((RNG.nextInt(distance) + 1) * ((RNG.nextBoolean())? +1 : -1));
      vDelta = ((RNG.nextInt(distance) + 1) * ((RNG.nextBoolean())? +1 : -1));
      testH = hBounds ( h + hDelta );     // on a torodial grid so wrap-around.
      testV = vBounds ( v + vDelta );     // on a torodial grid so wrap-around.

      // ------------------------------- QUERY --------------------------------
      found = ecoSys.getBiomeName (testH, testV).equals ("water");
      // ------------------------------- QUERY --------------------------------

      done = ( ( ++counter < cardinality ) ? false : true );
    }
    while ( !done && !found );

    if ( found )
    {
      Point p = new Point (testH, testV);              // a point in the water.
      p = findShallowWater (ecoSys, self.getCurrentLoc(), p); // shallow water.
      this.reqX = p.x;                           // assign to the transfer reg.
      this.reqY = p.y;                           // assign to the transfer reg.
    }
    else
    { // was tested and found not to be water so just move.
      this.reqX = testH;                         // assign to the transfer reg.
      this.reqY = testV;                         // assign to the transfer reg.
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method examines the cells adjacent to the current cell.  The
      goal is to identify a cell that satisfies an arbitrary goal.  In
      random order and constrained by the cardinality of the set enumerated
      by count of cells in the Moore neighborhood around this cell (a Moore
      distance set by parameter), the cells are randomly queried.  This
      is is a first fit algorithm.  The first cell that satisfies the query
      will be selected for recommendation.  If no cell is recommended the
      Class request registers (reqX and Y) will not changed.  If there is a
      recommendation, the resultant XY "move to" combination location will
      be put into the transfer registers.*/

  public boolean searchByFoodPreference (Hominid self, Ecology ecoSys,
                                         int distance)
  {

    int h = self.getCurrentLoc().x;                 // where self is in X-axis.
    int v = self.getCurrentLoc().y;                 // where self is in y-axis.
    setCurrentXY(h, v); // assume stay in the current location, load registers.

    String remoteForagable = "";          // this is the food in adjacent cell.

    boolean done = false;                    // to begin, the work is not done.
    boolean found = false;              // to begin, the solution is not found.

    int hDelta = 0;      // a distance from the location of self on the X axis.
    int vDelta = 0;      // a distance from the location of self on the Y axis.
    int testH = 0;                       // a value on the X axis to be tested.
    int testV = 0;                       // a value on the Y axis to be tested.

    int side = (2 * distance) + 1; // length of the search box side, an odd nr.
    int cardinality = (side * side) - 1;  // cardinality of adjacent Moore set.
    int counter = 0;       // a counter to compare to the cardinality variable.
    do
    {
      hDelta = ((RNG.nextInt(distance) + 1) * ((RNG.nextBoolean())? +1 : -1));
      vDelta = ((RNG.nextInt(distance) + 1) * ((RNG.nextBoolean())? +1 : -1));
      testH = hBounds ( h + hDelta );     // on a torodial grid so wrap-around.
      testV = vBounds ( v + vDelta );     // on a torodial grid so wrap-around.

      // ------------------------------- QUERY --------------------------------
      remoteForagable = ecoSys.getForageOption (testH, testV);
      found = ( remoteForagable.equals(self.getPreferredPrey()));
      // ------------------------------- QUERY --------------------------------

      done = ( ( ++counter < cardinality ) ? false : true );
    }
    while ( !done && !found );

    if ( found )
    {
      // test if is "water" then constrain movement if it is "into" the water.
      if (ecoSys.getBiomeName (testH, testV).equals("water"))
      {
        Point p = new Point (testH, testV);       // a point next to the water.
        p = findWatersEdge (ecoSys, self.getCurrentLoc(), p);  // water's edge.
        this.reqX = p.x;                         // assign to the transfer reg.
        this.reqY = p.y;                         // assign to the transfer reg.
      }
      else
      { // was tested and found not to be water so just move.
        this.reqX = testH;                       // assign to the transfer reg.
        this.reqY = testV;                       // assign to the transfer reg.
      }
    }

    return found;      // does a source of more food *AND* the preferred exist?

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method examines the cells adjacent to the current cell.  The
      goal is to identify a cell that satisfies an arbitrary goal.  In
      random order and constrained by the cardinality of the set enumerated
      by count of cells in the Moore neighborhood around this cell (a Moore
      distance set by parameter), the cells are randomly queried.  This
      is is a first fit algorithm.  The first cell that satisfies the query
      will be selected for recommendation.  If no cell is recommended the
      Class request registers (reqX and Y) will not changed.  If there is a
      recommendation, the resultant XY "move to" combination location will
      be put into the transfer registers.*/

  public void searchByFoodProbability (Hominid self, Ecology ecoSys,
                                       int distance)
  {

    int h = self.getCurrentLoc().x;                 // where self is in X-axis.
    int v = self.getCurrentLoc().y;                 // where self is in y-axis.
    setCurrentXY(h, v); // assume stay in the current location, load registers.

    float localForageProbability = 0.0F; // chance of this food source locally.
    float remoteProbability = 0.0F;                 // chance in adjacent cell.

    localForageProbability = ecoSys.getForageProbability(h, v);

    boolean done = false;                    // to begin, the work is not done.
    boolean found = false;              // to begin, the solution is not found.

    int hDelta = 0;      // a distance from the location of self on the X axis.
    int vDelta = 0;      // a distance from the location of self on the Y axis.
    int testH = 0;                       // a value on the X axis to be tested.
    int testV = 0;                       // a value on the Y axis to be tested.

    int side = (2 * distance) + 1; // length of the search box side, an odd nr.
    int cardinality = (side * side) - 1;  // cardinality of adjacent Moore set.
    int counter = 0;       // a counter to compare to the cardinality variable.
    do
    {
      hDelta = ((RNG.nextInt(distance) + 1) * ((RNG.nextBoolean())? +1 : -1));
      vDelta = ((RNG.nextInt(distance) + 1) * ((RNG.nextBoolean())? +1 : -1));
      testH = hBounds ( h + hDelta );    // on a torodial grid so wrap-around.
      testV = vBounds ( v + vDelta );    // on a torodial grid so wrap-around.

      // ------------------------------ QUERY --------------------------------
      remoteProbability = ecoSys.getForageProbability (testH, testV);
      found = (remoteProbability > localForageProbability);
      // ------------------------------ QUERY --------------------------------

      done = ( ( ++counter < cardinality ) ? false : true );
    }
    while ( !done && !found );

    if ( found )
    {
      // test if is "water" then constrain movement if it is "into" the water.
      if (ecoSys.getBiomeName (testH, testV).equals("water"))
      {
        Point p = new Point (testH, testV);       // a point next to the water.
        p = findWatersEdge (ecoSys, self.getCurrentLoc(), p);  // water's edge.
        this.reqX = p.x;                         // assign to the transfer reg.
        this.reqY = p.y;                         // assign to the transfer reg.
      }
      else
      { // was tested and found not to be water so just move.
        this.reqX = testH;                       // assign to the transfer reg.
        this.reqY = testV;                       // assign to the transfer reg.
      }
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method sets copies the parameters h and v into local reqX & reqY.*/

  //private void setCurrentXY(int h, int v)
  public void setCurrentXY(int h, int v)
  {

    this.reqX = h;
    this.reqY = v;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets the parameter reqX.*/

  public void setReqX(int newX)
  {

    if (newX < 0)
    {
      System.out.println("Error mode: Heuristics.getReqX() - 0");
      System.out.println("            Did you reset/adjust ModelConstraints?");
    }
    if (newX > (hLimit - 1))
    {
      System.out.println("Error mode: Heuristics.getReqX() - max");
      System.out.println("            Did you reset/adjust ModelConstraints?");
    }

    this.reqX = newX;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method gets the parameter reqY.*/

  public void setReqY(int newY)
  {

    if (newY < 0)
    {
      System.out.println("Error mode: Heuristics.getReqY() - 0");
      System.out.println("            Did you reset/adjust ModelConstraints?");
    }
    if (newY > (vLimit - 1))
    {
      System.out.println("Error mode: Heuristics.getReqY() - max");
      System.out.println("            Did you reset/adjust ModelConstraints?");
    }

    this.reqY = newY;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This constrains vertical movement onto a tordial surface. There are
      three possible states: too far above, too far below, or on grid.*/

  public int vBounds (int possibleV)
  {

    if (possibleV < 0)                                        // too far above.
      return vLimit + possibleV;                  // circle in from the bottom.

    if (possibleV < vLimit)                          // if less than the limit.
      return possibleV;                            // no problem, stay on grid.

    if (possibleV >= vLimit)                               // if too far below.
      return possibleV - vLimit;                     // circle in from the top.

    System.out.println ("ERROR MODE: HAL.vBounds testing " + possibleV);
    return -42;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does something.*/

  public void template()
  {

    //add code here.

  } // End method.
  //***************************************************************************

} // End of class.
