//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Clans.java                                                       *
//* CLASS    HominidKit.java (Hominid transported (transportable) objects)    *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     06/06/2015                                                       *
//* REVISION X.0.1.2                                                          *
//* BYTES    00,000                                                           *
//* TIME     15:55                                                            *
//* PURPOSE  This provides objects a migratory individual might need.         *
//* METHOD   Each instance of a class is a stand-alone code object.           *
//* DETAILS  A Java document exists that outlines class operation.  Other     *
//*          documentation in the form of programmer notes apply.             *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root.hominini;                                       // home directory.

import java.util.ArrayList;                                   // Java language.
import java.util.Random;                                      //         ditto.

import root.hominini.HominidKit;                       // program dependencies.
import root.ModelConstraints;                          //                ditto.

//
//*****************************************************************************
// Hominid Kit

public class HominidKit implements ModelConstraints
{

  //***************************************************************************
  // Class constants, variables, and objects.

  // Constants.
  final float   SUPPLY_CARRYING_LIMIT = 0.1F;  // max supply carrying capacity.
  final float   SUPPLY_START_AMOUNT = 0.100F; // amount of supplies start with.
  final float   SUPPLY_DEPLETION_RATE = 0.010F;    // amount needed to do work.
  final int     FLASK_TIME_LIMIT = 14; // max days until water container leaks.
  final float   MAX_SURVIVAL_TECH = 0.955F;       // max survival tech on hand.

  final float   HOMINID_NOT_IMPERILED = 0.0F;       // indiv absolutely secure.
  final float   HOMINID_PERIL_RATE = 0.01F;  // rate an imperiled hominid dies.
  final float   INDIV_IS_IMPERILED = 1.0F;      // indiv absolutely threatened.

  // Variables
  // NOTE:  water is measured in gallons and food in pounds.
  private float waterOnHand = 0.0F;       // the current, on hand water amount.
  private float waterCapacity = 0.0F;       // water carrying capacity in gals.
  private int   currentFlaskTime = 0;       // current time on the flask skins.
  private float foodOnHand_F = 0.0F;    // current, on hand food amt in pounds.
  private float foodOnHand_M = 0.0F;    // current, on hand food amt in pounds.
  private float foodCapacity_F = 0.0F;     // food carrying capacity in pounds.
  private float foodCapacity_M = 0.0F;     // food carrying capacity in pounds.
  private float suppliesOnHand = 0.0F; // the current, on-hand supplies amount.
  private boolean hasTech = false;            // current boolean state of tool.

  private float stateOfPeril = 0.0F;            // the hominid is in no danger.

  private static long rng_SEED = 0L;            // this is the GLOBAL RNG seed.

  // Object.
  private static Random RNG;                 // pointer to standard RNG object.

  //***************************************************************************
  /** this is the class constructor.*/

  public HominidKit (long newSeed)
  {

    rng_SEED = newSeed;                            // copy the GLOBAL RNG seed.
    RNG = new Random(rng_SEED);                 // instantiate RNG with a seed.

    initialize ();                                    // initialize the cohort.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method returns the state of repair for the water containers.*/

  private int getFlaskTime()
  {

    return this.currentFlaskTime;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method sets the female hominid's food carrying capacity.*/

  private void setFoodCapacity_F(float newFoodCapacity)
  {

    this.foodCapacity_F = newFoodCapacity;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method sets the male hominid's food carrying capacity.*/

  private void setFoodCapacity_M(float newFoodCapacity)
  {

    this.foodCapacity_M = newFoodCapacity;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method returns the hominid's water carrying capacity.*/

  private float getFlaskCapacity()
  {

    return this.waterCapacity;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method returns the female hominid's food carrying capacity.*/

  private float getFoodCapacity_F()
  {

    return this.foodCapacity_F;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method returns the male hominid's food carrying capacity.*/

  private float getFoodCapacity_M()
  {

    return this.foodCapacity_M;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method returns the female hominid's food on hand.*/

  public float getFoodOnHand_F()
  {

    return this.foodOnHand_F;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method returns the male hominid's food on hand.*/

  public float getFoodOnHand_M()
  {

    return this.foodOnHand_M;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method returns the CONSTANT of hominid's peril status limit.*/

  public float getINDIV_IS_IMPERILED()
  {

    return INDIV_IS_IMPERILED;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method returns the hominid's peril status.*/

  public String getPerilReport()
  {

    if (getPerilStatus() == HOMINID_NOT_IMPERILED)
      return "NonImperiled";
    else if (getPerilStatus() == INDIV_IS_IMPERILED)
      return "Dead";
    else
      return "Imperiled";

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method returns the hominid's peril status. 0.0 is completely safe.*/

  public float getPerilStatus()
  {

    return this.stateOfPeril;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method returns the amount of supply material with the hominid.*/

  public float getSuppliesOnHand()
  {

    return this.suppliesOnHand;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method returns the hominid's water on hand.*/

  public float getWaterOnHand()
  {

    return this.waterOnHand;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method updates the state of repair for the water containers.*/

  private void incrementFlaskTime ()
  {

    int tempTime = getFlaskTime();
    this.setFlaskTime(++tempTime);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method works in cooperation with the Class constructor to setup
      the runtime environment.*/

  public void initialize ()
  {

    setPerilStatus(HOMINID_NOT_IMPERILED);    // hominid notimperiled at start.
    setTechStatus(true);                      // Assume hominid has their tech.

    // set maximum water carrying capacity and the amount on hand.
    setWaterCapacity(WATER_CARRYING_LIMIT);
    setWaterOnHand(WATER_START_AMOUNT);

    // set maximum food carrying capacity and the amount on hand.
    setFoodCapacity_F(FOOD_CARRYING_LIMIT_F);
    setFoodCapacity_M(FOOD_CARRYING_LIMIT_M);
    setFoodOnHand_F(FOOD_START_AMOUNT);
    setFoodOnHand_M(FOOD_START_AMOUNT);

    // sets the current time on the water flasks.
    setFlaskTime(FLASK_TIME_LIMIT);

    setSuppliesOnHand(SUPPLY_START_AMOUNT);            // an arbitrary  amount.

    //kitReport();                                               // DIAGNOSTIC.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this returns a report on this hominid's kit as an object.*/

  public void kitReport (Hominid hominid)
  {

    long agentID = hominid.getAgentID();
    HominidKit kit = hominid.kit;
    float foodAmount = (hominid.bio.getSex().equals("M") ?
                        kit.getFoodOnHand_M() :
                        kit.getFoodOnHand_F());

    System.out.println("#################################################");
    System.out.println("HominidKit.kitReport()\n");

    System.out.println("For agent: " + agentID);
    System.out.println("Water on-hand (gal):   " + kit.getWaterOnHand());
    System.out.println("Food on-hand (lb):     " + foodAmount);
    System.out.println("Supplies on-hand (lb): " + kit.getSuppliesOnHand());
    System.out.println("Peril status:          " + kit.getPerilStatus());
    System.out.println("Kit supply levels:     " + kit.getPerilReport());
    System.out.println("Survival tech on-hand: " + kit.getTechStatus());
    System.out.println("\n");

    System.out.println("#################################################");

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method sets the state of repair for the water containers.*/

  private void setFlaskTime(int newFlaskTime)
  {

    this.currentFlaskTime = newFlaskTime;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method sets the female hominid's food on hand.*/

  public void setFoodOnHand_F(float newFoodOnHand)
  {

    this.foodOnHand_F = newFoodOnHand;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method sets the male hominid's food on hand.*/

  public void setFoodOnHand_M(float newFoodOnHand)
  {

    this.foodOnHand_M = newFoodOnHand;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method sets the hominid's water carrying capacity.*/

  private void setWaterCapacity(float newWaterSource)
  {

    this.waterCapacity = newWaterSource;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method sets the amount of supply material with the hominid.*/

  public void setSuppliesOnHand(float newAmountOfSupplies)
  {

    this.suppliesOnHand = newAmountOfSupplies;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /**This method returns the hominid's tool status.*/

  private boolean getTechStatus()
  {

    return this.hasTech;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method allows setting the hominid's tool status.*/

  private void setTechStatus(boolean setHasTechStatus)
  {

    this.hasTech = setHasTechStatus;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method allows external health improvements to made to the hominid's
      peril status.  It also restarts the water container time counter with
      a random value.*/

  private void repairWaterContainer()
  {

    // resets the current water container time to something less than nothing.
    setFlaskTime((int)(RNG.nextFloat() * (float) FLASK_TIME_LIMIT));
    setPerilStatus(HOMINID_NOT_IMPERILED);     // Partially reprovision cohort.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method allows setting the hominid's peril status.  0.0 is secure.
      1.0 is completely imperiled.  It limits the status maximum and
      minimums.*/

  public void setPerilStatus(float newPerilStatus)
  {

    this.stateOfPeril = newPerilStatus;

    if (this.stateOfPeril > INDIV_IS_IMPERILED)
      this.stateOfPeril = INDIV_IS_IMPERILED;    // allow values up to the max.

    if (this.stateOfPeril < HOMINID_NOT_IMPERILED)
      this.stateOfPeril = HOMINID_NOT_IMPERILED;//allow values down to the min.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method sets the hominid's water on hand.*/

  public void setWaterOnHand(float newWaterOnHand)
  {

    this.waterOnHand = newWaterOnHand;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method tests for the likelihood of the hominid still having their
      most necessary survival technology(s).*/

  private void updateHasTool ()
  {

    // let a random number decide if the technology(s) are still available.
    if (RNG.nextFloat() < MAX_SURVIVAL_TECH)
      setTechStatus(true);                          // yup, got tech right now.
    else
      setTechStatus(false);                     // nope, got no tech right now.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method updates the kit food portion.  Only food kit is updated.*/

  public void updateKitFood (String sex, float foodFound)
  {

    // add to kit according to what has been provided: mind sex and limits.
    float tempFood = 0.0F;                              // foodFound is in lbs.

    if (sex.equals("M"))
    {
      tempFood = foodFound + getFoodOnHand_M();         // foodFound is in lbs.
      if (tempFood > FOOD_CARRYING_LIMIT_M)  // only hand-to-mouth amount (lbs).
        setFoodOnHand_M(FOOD_CARRYING_LIMIT_M);//hld only the food you can eat.
      else
        setFoodOnHand_M(tempFood);          // less than you want, but keep it.
    }
    else
    {
      tempFood = foodFound + getFoodOnHand_F();         // foodFound is in lbs.
      if (tempFood > FOOD_CARRYING_LIMIT_F) // only hand-to-mouth amount (lbs).
        setFoodOnHand_F(FOOD_CARRYING_LIMIT_F);//hld only the food you can eat.
      else
        setFoodOnHand_F(tempFood);          // less than you want, but keep it.
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method updates the kit supplies.  Only supplies are updated.*/

  public void updateKitSupplies (float supplies)
  {

    // add to kit according to what has been provided and mind the limits.
    float tempSupplies = supplies + getSuppliesOnHand();
    if (tempSupplies > SUPPLY_CARRYING_LIMIT)
      setSuppliesOnHand(SUPPLY_CARRYING_LIMIT);            // restore supplies.
    else
      setSuppliesOnHand(tempSupplies);

    // does hominid still have their primary technologies?
    updateHasTool();
    if (! getTechStatus())             // check Hominid possession of kit tech.
    {
      float tempStatus = getPerilStatus() + HOMINID_PERIL_RATE;
      setPerilStatus (tempStatus);
    }

    // does the cohort still have its primary means of transporting water?
    this.incrementFlaskTime();        // update wear-time counter on the flask.
    if (getFlaskTime() >= FLASK_TIME_LIMIT)           // check wear-time limit.
    {
      // are supply reserves enough to make repairs to water container?
      if (getSuppliesOnHand() >= SUPPLY_DEPLETION_RATE)
      {
        float tempStatus = getSuppliesOnHand();
        setSuppliesOnHand(tempStatus - SUPPLY_DEPLETION_RATE);
        repairWaterContainer();          // Great news.  Repairs were possible.
      }
      else  // oops. this is bad....
      {
        float tempStatus = getPerilStatus();
        setPerilStatus (tempStatus + HOMINID_PERIL_RATE);
      }
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** this method updates the kit water portion.  Only water kit is updated.*/

  public void updateKitWater (float waterFound)
  {

    // add to kit according to what has been provided and mind the limits.
    float tempWater = waterFound + getWaterOnHand();       // water is in gals.

    if (tempWater > WATER_CARRYING_LIMIT)   // only hand-to-mouth amount (gal).
      setWaterOnHand(WATER_CARRYING_LIMIT);  // hold only  water you can drink.
    else
      setWaterOnHand(tempWater);            // less than you want, but keep it.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does something.*/

  public void template()
  {

    //add code here.

  } // End method.
  //***************************************************************************

} // End of class.
