//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Clans.java                                                       *
//* CLASS    SeedGroup.java (starts the simulated Hominid population)         *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     06/06/2015                                                       *
//* REVISION X.0.1.2                                                          *
//* BYTES    00,000                                                           *
//* TIME     15:55                                                            *
//* PURPOSE  This object instantiates a seed group of simulated Hominids.     *
//* METHOD   Each instance of a class is a stand-alone code object.           *
//* DETAILS  We interpret the metaphor of a cohort to include individuals     *
//*          who are colocated spatially and share common activities.         *
//*          A Java document exists that outlines class operation.  Other     *
//*          documentation in the form of programmer notes apply.             *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root.hominini;                                       // home directory.

import java.util.ArrayList;                                   // Java language.
import java.util.Random;                                      //         ditto.
import java.awt.Point;                                        //         ditto.

import root.hominini.Hominid;                          // program dependencies.
import root.events.LifeExperiences;                    //                ditto.
import root.rng.TrueRNG;                               //                ditto.
import root.ModelConstraints;                          //                ditto.

//
//*****************************************************************************
// Seed Group

public class SeedGroup implements ModelConstraints
{

  //
  //***************************************************************************
  // Class variables and objects.

  // Constants.
  private static Point [] fixedGridSet =  {new Point ( 66, 218),
                                           new Point (327, 184)};

  // Variables.
  private static int hLimit = 0;                    // horizontal grid surface.
  private static int vLimit = 0;                      // vertical grid surface.
  private long rng_Seed = 0;                    // this is the GLOBAL RNG seed.
  private int agentsAlive = 0;              // after seed-group forms holds nr.
  private int totalAgentsSimulated = 0;     // after seed-group forms holds nr.

  // Objects.
  private LifeExperiences life = null;          // the daily lives of hominids.
  private static Random RNG = null;          // pointer to standard RNG object.
  private static TrueRNG trueRNG = null;      // pointer to TrueRNG USB device.

  //***************************************************************************
  /** This is the class constructor for the SeedGroup.*/

  public SeedGroup (long newSeed, TrueRNG devicePntr,
                    int horizontalLimit, int verticalLimit,
                    LifeExperiences virtualLife)
  {

    this.rng_Seed = newSeed;            // make a local copy of the fixed seed.

    // if the TrueRNG USB device is available (ie devicePntr is not null) then,
    // the seed used by RNG will be truly random.  Otherwise, a fixed value of
    // newSeed will be used and all stochastic patterns will be deterministic.
    if (devicePntr == null)
      RNG = new Random(newSeed);  // instantiate an RNG object with fixed seed.
    else
    {
      trueRNG = devicePntr;                // make a local copy of the pointer.
      RNG = new Random (devicePntr.getNextRandom()); // create with real rndom.
    }

    this.hLimit = horizontalLimit;                        // take a local copy.
    this.vLimit = verticalLimit;                          // take a local copy.
    this.life = virtualLife;  // pntr back into life experiences virtual world.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the horizontal coordinate of a 2-D array where
      the hLimit dimension of the array is known GLOBALLY.*/

  public int findHCoordinate (int uRDist [], int cellNr)
  {

    // all row positions are modulus functions of the width of the matrix.
    return ((uRDist [cellNr]) % this.hLimit);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns the vertical coordinate of a 2-D array where
      the hLimit dimension of the array is known GLOBALLY.*/

  public int findVCoordinate (int uRDist [], int cellNr)
  {
    int v = 0;

    // test if value is in the first row, else must be a later row.
    if ( uRDist [cellNr] < this.hLimit)
      v = 0;
    else
      v = uRDist [cellNr] / this.hLimit;

    return v;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method generates and returns a set of pseudo-random, uniformly
      distributed values.  NOTE findHCoordinate() and findVCoordinate() are
      helpful in mapping these values onto a 2-D coordinate matrix.*/

  public int [] getUniformRandom (int galoisFieldSize)
  {

    final byte EMPTY = (byte) 0x00;                        // symbol for EMPTY.
    final byte FULL =  (byte) 0x01;                         // symbol for FULL.

    int [] uniformRandomList = new int[galoisFieldSize];   // the return field.
    byte [] checkList = new byte [galoisFieldSize];       // used for checking.

    for (int z = 0; z < galoisFieldSize; z++)
      checkList [z] = EMPTY;     // fill the backstop list with NOT USED flags.

    int sample = 0;                               // this is the random sample.
    int g = 0;
    {

      do
      { // continue looping until every EMPTY checkList cell is filled.

        sample = RNG.nextInt (galoisFieldSize);
        if (checkList [sample] == EMPTY)
        {
          uniformRandomList [g++] = sample;
          checkList [sample] = FULL;
        }

      } while (g < galoisFieldSize);

    }

    return uniformRandomList;                // return the uniform random list.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** WARNING:  this is a long method.

      This method generates groups of hominid agents as spatially co-located
      cohorts.  The size, number, and spatial location of the cohorts
      are set by the value of the parameters supplied to the method when
      it is called.  An seed group is returned as ArrayList of Hominids.
      Within each cohort a single Alpha will have been set and each member
      of the cohort will share group self-awareness.*/

  public ArrayList <Hominid> gridStart ( int epoch,
                                         int cohortSize, int nrOfCohorts,
                                         LifeExperiences life)
  {

    System.out.println("Creating seed population(s).");     // announce status.

    // establish the initial agent culture in the simulation by generating
    // the required number of mixed age, probabilistically gendered, count-
    // constrained, individual-centric, hominid cohorts.
    Hominid p = null;             // the person that is the focus of attention.
    ArrayList <Hominid> cohorts = new ArrayList <Hominid> ();   // all cohorts.
    ArrayList <Hominid> cohort = new ArrayList <Hominid> ();     // one cohort.
    int maxHomCount = nrOfCohorts * cohortSize;  // calc max nr hominid agents.

    int [] urDistribution = null;                         // a UR distribution.
    int galoisSetCardinality = this.hLimit * this.vLimit;
    urDistribution = getUniformRandom(galoisSetCardinality);   // one for each.

    int homPlaced = 0;       // the overall count of hominid placed in cohorts.
    int cohortsPlaced = 0;              // the overall count of cohorts placed.
    int cohortSizeReq = 0;      // the number of homis required in each cohort.

    // work across each individual cohort, then loop and do another until done.
    cohort = new ArrayList <Hominid> ();
    do
    {

      Point pnt = new Point();                 // a cohort will form here next.
      if (FIXED_GRID)
      {
        pnt = fixedGridSet [cohortsPlaced];    // used for calibration testing.
      }
      else
      {
        pnt.x = findHCoordinate (urDistribution, cohortsPlaced);    // dynamic.
        pnt.y = findVCoordinate (urDistribution, cohortsPlaced);    // dynamic.
      }

      cohortSizeReq = cohortSize;
      if (( homPlaced + cohortSizeReq ) > maxHomCount)// test and limit number.
        cohortSizeReq =  maxHomCount - homPlaced;    // reset to the new limit.

      for (int i = 0; i < cohortSizeReq; i++)
      {
        p = new Hominid (RNG.nextLong(), trueRNG, epoch, pnt,
                this.hLimit, this.vLimit, this.life);        // create & place.
        cohort.add(p);                       // now add them to the new cohort.
        homPlaced += 1;          // update the count of hominid placed overall.
      }
      cohortsPlaced += 1;        // update the count of cohorts placed overall.
   // life.makeAlpha(cohort);       // make one member of this cohort an alpha.
      informGroupsOf (cohort);                  // start cohort self-awareness.

      cohorts.addAll(cohort); // add this cohort to the master list of cohorts.
      cohort.clear();     // the cohort is now complete so clear the variables.

    } while (homPlaced < maxHomCount);    // continue until all are in cohorts.

    cohort = null;                                        // return the memory.
    p = null;                                             // return the memory.

    System.out.println("Number of cohort groups:  " + cohortsPlaced + ".");
    System.out.println("Number of agents started: " + homPlaced + ".");
    System.out.println("Cohort setup is now complete.\n");  // announce status.

    // ------------------------------------------------------------------------
    // update all clerical accounts associated with the addition of agents.
    this.agentsAlive = cohorts.size();                    // update the record.
    this.totalAgentsSimulated = cohorts.size();           // update the record.
    // ------------------------------------------------------------------------

    // DIAGNOSTIC.
    // this will display the entire seedGroup cohort in one long sequence.
    //for (int i = 0; i < cohorts.size(); i++)
    //  cohorts.get(i).report();
    // DIAGNOSTIC.

    return cohorts;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method updates (on a cohort basis) the individual group list(s)
      of the agents identified in the localCohort parameter.*/

  public static void informGroupsOf (ArrayList <Hominid> localCohort)
  {

    Hominid member = null;
    Hominid hominid = null;

    int limit = localCohort.size ();
    for (int i = 0; i < limit; i++)
    {
      member = localCohort.get (i);
      for (int j = 0; j < limit; j++)
      {
        hominid = localCohort.get (j);
        if (hominid != member)
          hominid.kin.updateGroup (member);    // FILTER:  only deltas allowed.
      }
    }

    member = null;                // return the memory to the operating system.
    hominid = null;               // return the memory to the operating system.

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns a number describing the agents created during the
      seed-group formation process.*/

  public int getAgentsAlive ()
  {

    return this.agentsAlive;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method returns a number describing the agents created during the
      seed-group formation process.*/

  public int getTotalAgentsSimulated ()
  {

    return this.totalAgentsSimulated;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method does something.*/

  public void template ()
  {

    // add code here.

  } // End method.
  //***************************************************************************

}
