//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      Clans.java                                                       *
//* CLASS    TrueRNG.java. (True Random Number Generator)                     *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     06/06/2015                                                       *
//* REVISION X.0.1.2                                                          *
//* BYTES    00,000                                                           *
//* TIME     15:55                                                            *
//* PURPOSE  To setup, control, and interface with a TrueRNG USB device.      *
//* METHOD   Uses serial communications to operate the TrueRNG device.        *
//* DETAILS  RXTX libraries must be in compile and native paths.  Also,       *
//*          sub-directories ./gnu.io.* must exist with four imports          *
//*          available.  See the import list below.                           *
//* ENVIRON  Developed on Intel Core-2 Quad Q8200 2.33 GHz CPU, Nvidia-based  *
//*          PCI-X graphics acceleration.  The OS was MS Windows 7 Ultimate.  *
//*          Java IDE was JGrasp version 2.0.0_08.  The Java compiler was     *
//*          version 1.8.0.  Verification tests were run in Java JRE          *
//*          build 1.8.0-b132 with client VM build 14.3-b01.                  *
//*                                                                           *
//*****************************************************************************
//

package root.rng;                                            // home directory.

import java.io.BufferedReader;                                // Java language.
import java.io.FileNotFoundException;                         //         ditto.
import java.io.FileOutputStream;                              //         ditto.
import java.io.InputStreamReader;                             //         ditto.
import java.io.IOException;                                   //         ditto.
import java.util.Enumeration;                                 //         ditto.
import java.util.Random;                                      //         ditto.

import gnu.io.CommPortIdentifier;                 // hardware dependency trees.
import gnu.io.PortInUseException;                             //         ditto.
import gnu.io.SerialPort;                                     //         ditto.
import gnu.io.UnsupportedCommOperationException;              //         ditto.

import root.ModelConstraints;                              // program routines.

//
//*****************************************************************************
// TrueRNG

public class TrueRNG implements ModelConstraints
{

  //
  //***************************************************************************
  // Class constants, variables, and objects.

  // Constants.
  final static int NR_OF_BYTES = 8;  // factory Random uses 6-byte 48-bit seed.
  final static int RESET_LIMIT = 0;    // works with resetCounter as its limit.
  final static int COUNTER_MAX_SPIN = 17;        // the max resetCounter limit.
  final static int COUNTER_MIN_SPIN = 9;         // the min resetCounter limit.

  // Variables.
  private static String trueRNGPort = "COM5";  // this is the current COM port.
  private static SerialPort rngPort = null;  // this is the main COM port link.
  private static BufferedReader bffrdReader = null;  // this connects COM port.
  private static boolean startedOK = true;          // a status reporting flag.
  private static int resetCounter;        // counts down from X to RESET_LIMIT.

  // Objects.
  private static RNGKernel rngThread = null;  //  pointer to the private class.

  //***************************************************************************
  /** This is a private class constructor.*/

  private static class RNGKernel extends Thread
  {

    final static int BITS_PER_BYTE = 8;                     // 8-bits per byte.

    private boolean terminateRun = false;   // controls the repeating run loop.
    private long nextRandom = 0L;          // the next available random number.
    private int bufferLength = BITS_PER_BYTE * NR_OF_BYTES; // total bit count.
    private BufferedReader bReader = null;       // connection to the COM port.
    private Random RNG = new Random(System.nanoTime());   // factory pseudoRNG.

    // this takes an array of chars and returns a converted 64-bit long.
    private long convertToLong (char [] arrayOfChars) {
      String stringOfChars = new String (arrayOfChars);

      long result = 0L;
      long factor = 1L;
      long sample = 0L;
      for (int index = 0; index < stringOfChars.length(); index++)
      {
        sample = (long) ((byte) stringOfChars.codePointAt (index)); // nipNtuc.
        result = result + Math.abs (sample * factor);
        factor *= 256L;
      }
      return result;
    }

    // this takes an array of chars and returns a converted 32-bit int.
    private long convertToInt (char [] arrayOfChars) {
      String stringOfChars = new String (arrayOfChars);

      int result = 0;
      int factor = 1;
      int sample = 0;
      for (int index = 0; index < stringOfChars.length(); index++)
      {
        sample = (int) ((byte) stringOfChars.codePointAt (index));  // nipNtuc.
        result = result + Math.abs (sample * factor);
        factor *= 256;
      }
      return result;
    }

    // this is the accessor pickup of the next random 64-bit long.
    private synchronized long getNextRandom () {
      if (resetCounter == RESET_LIMIT)
      {
        RNG.setSeed(nextRandom);
        resetCounter = COUNTER_MIN_SPIN + RNG.nextInt (COUNTER_MAX_SPIN);
      }
      // we're always waiting for COM port to respond, so do a little shuffle.
      nextRandom = Long.rotateLeft (nextRandom,
                                 (1 + RNG.nextInt (bufferLength - 1)));
      return nextRandom;
    }

    // this connects to the COM port via the bReader and returns a char array.
    private char [] readChar () {
      char[] charBlock = new char[NR_OF_BYTES];
      try
      {
        do {} while (NR_OF_BYTES == bReader.read(charBlock));
      }
      catch (IOException ioe)
      {
        System.out.println("TrueRNG.RNGKernel.readChar terminated.");
      }
      return charBlock;
    }

    // this is an externally writable connector to the COM port.
    private void setBufferedReader (BufferedReader bffrdReader) {
      bReader = bffrdReader;
    }

    // this is an externally writable control on the main loop.
    private synchronized void setTerminate (boolean newState) {
      terminateRun = newState;
    }

    // this is the main threaded loop.  It directs the port reading, data
    // conversion activity, and aperiodic reset of the local "fast" java RNG.
    public void run () {
      do
      {
        nextRandom = convertToLong (readChar ());
      } while (!terminateRun);
    }

  }
  //***************************************************************************

  //***************************************************************************
  /** This is the class constructor.*/

  public TrueRNG (String portToUse)
  {

    trueRNGPort = portToUse;

  }
  //***************************************************************************

  //***************************************************************************
  /** This method provides an external accessor for the current random.*/

  public synchronized long getNextRandom ()
  {

    return rngThread.getNextRandom ();

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method will initialize the named COM port.*/

  public SerialPort initializePort (SerialPort rngPort)
  {

    SerialPort port = rngPort;

    try
    {
      port.setSerialPortParams(115200,
                               SerialPort.DATABITS_8,
                               SerialPort.STOPBITS_1,
                               SerialPort.PARITY_NONE);

      port.setFlowControlMode(SerialPort.FLOWCONTROL_NONE);
      port.setDTR(true);
      port.setRTS(true);
      System.out.print(", initialized");
    }
    catch (UnsupportedCommOperationException ucoe)
    {
      System.out.println(", and failed.");
      System.out.println("ERROR: TrueRNG.SerialPort.");
      startedOK = false;
    }

    return port;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method opens an identified available port.*/

  public SerialPort openPort (CommPortIdentifier portID,
                                     String requestingApp)
  {

    SerialPort port = null;

    try
    {
      try
      {
        port = (SerialPort) portID.open (requestingApp, 1000 );
      }
      catch (NullPointerException npe)
      {
        System.out.print ("ERROR:TrueRNG.SerialPort check USB device.");
      }
      System.out.print (", opened");
    }
    catch(PortInUseException e)
    {
      System.out.println(" but failed.");
      System.err.println("ERROR: " + portID.getName() + " failed to open.");
      startedOK = false;
    }

    return port;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method identifies the named port if it exists.*/

  public CommPortIdentifier portEnumerator (String portNamed)
  {

    Enumeration portIdentifiers = CommPortIdentifier.getPortIdentifiers();

    if (!portIdentifiers.hasMoreElements())
    {
      System.out.println("No serial ports found.");
      startedOK = false;
    }

    CommPortIdentifier portId = null;
    CommPortIdentifier pid = null;

    while (portIdentifiers.hasMoreElements())
    {

      pid = (CommPortIdentifier) portIdentifiers.nextElement();
      if ((pid.getPortType() == CommPortIdentifier.PORT_SERIAL) &&
        pid.getName().equals(portNamed))
      {
        portId = pid;
        break;
      }
    }

    if (portId == null)
    {
      System.err.println("Could not find serial port " + portNamed);
      startedOK = false;
    }
    else
      System.out.print ("TrueRNG found on " + pid.getName());

    return portId;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is an external coordinator to setup a COM port.*/

  public boolean setupDevice ()
  {

    final String REQUESTING_APPLICATION = "dummy";

    CommPortIdentifier portID = portEnumerator(trueRNGPort);

    if (startedOK)
      rngPort = openPort(portID, REQUESTING_APPLICATION);

    if (startedOK)
      rngPort = initializePort(rngPort);

    if (startedOK)
    {
      setupReader (rngPort); // char
    }

    return startedOK;

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method uses the shared data buffer and connects it with the port.*/

  public void setupReader (SerialPort rngPort)
  {

    try
    {
      bffrdReader = new BufferedReader (
                    new InputStreamReader (rngPort.getInputStream ()));
      System.out.println (", and setup.");
    }
    catch (IOException e)
    {
      System.err.println ("ERROR: TrueRNG.setupReader.");
      bffrdReader = null;
      startedOK = false;
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method closes the output file image.*/

  public void shutdownOutputFile(FileOutputStream out)
  {

    if (out != null)
    {
      try
      {
        out.close();
      }
      catch (IOException ioe)
      {
        System.out.println("ERROR: TrueRNG.shutdownOutputFile.");
      }
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is an external coordinator that starts the TrueRNG device.*/

  public void startDevice()
  {

    final int MS_DELAY = 1000;                   // delay while TrueRNG starts.

    rngThread = new RNGKernel ();
    rngThread.setBufferedReader (bffrdReader);
    rngThread.start ();

    wait (MS_DELAY);

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is an external image of the device stopper.*/
  public void stopDevice ()
  {

    rngThread.setTerminate (true);
    do
    {
      // loop while waiting for the RNG thead to die.
    } while (rngThread.isAlive ());

    try
    {
      if (bffrdReader != null)
        bffrdReader.close();
    }
    catch (IOException ioe)
    {
      System.out.println ("ERROR: TrueRNG.stopDevice.");
    }

    if (rngPort != null)
    {
      rngPort.close();
    }

  } // End method.
  //***************************************************************************

  //***************************************************************************
  /** This method is prepared, millisecond-timed, in-line wait state.*/

  public void wait (int milliseconds)
  {

    try
    {
      Thread.sleep (milliseconds);
    }
    catch (InterruptedException ie)
    {
      System.out.println ("ERROR: TrueRNG.wait.");
    }

  } // End method.
  //***************************************************************************

}

